get_true_rnd:
  .286

  comment *
  TITLE
    get_true_rnd / RND.BIN (compiled code fragment)

    Public domain true random number generator.  Free for all users and
    programmers.  Use and/or modify at your own risk.

  AUTHOR
    Russell Leidich (CompuServe 71271,2471).

  DATE
    September 25, 1994

  IN
    AX = Randomness (from 1 to 4096).

  OUT
    AX = True random number (as close as it gets).

  ASSUMPTIONS
    IF is set.  (STI is in effect).  Runnable on a 286 or higher.

  VITAL REMARKS
    A randomness of 50 or so should be sufficient for most applications.  If
    you're really insecure, however, you can boost it up to 4096.  (The
    execution time of get_true_rnd is directly proportional to the randomness.)
    The randomness is a measure of the quality of the random number.  It is not
    a seed.  Faster CPUs will require higher randomness values.

    DO NOT attempt to covert this routine for 8086 use, as the analog chaos
    that it relies on is destroyed by the PC-XT 4.77 MHz oscillator.

    If it seems that timer IO is not occurring properly, add more IO delay (JMP
    $+2 -- see code below).  I have not encountered a machine that had a
    problem with this, but one might exist.

  BACKGROUND
    Encrpytion routines in particular, but also a wide variety of computer
    games, require random numbers.  Beyond that, they may require so-called
    "true" random numbers.  The distinction is that true random numbers are
    different every time they are chosen, unlike pseudo-random numbers, which
    are repeatable by resetting a "seed" -- a digital initial condition to a
    digital chaos generator.

    A pseudo-random number generator (PRNG) uses the seed to produce a sequence
    of seemingly-unrelated numbers from it.  Trouble is, most of these
    generators have statistical tendencies which restrict the effective
    universe of possible "random" numbers that they generate.  In fact, it can
    be proven that any such pseudo-random number generator must eventually
    repeat itself, regardless of how many (but finite) binary operations it
    performs.

    Suppose you use a PRNG to select passwords for a user to choose from.  Your
    passwords might be 64 bits long, which leads you to believe that there over
    nine quintillion possible passwords that a hacker would have to try (on
    average) in order to obtain the user's private information.  However, if
    you are generating these passwords for a PRNG that can only produce a
    million unique passwords, then you are far less secure than you thought!

    One approach to true random number generation is to get statistics on
    everything in the system -- from keyboard IO to disk space, etc.  This is a
    cumbersome, time-consuming, and quite frankly rediculous method which may
    still involve statistical tendencies simply because the vast majority of
    information on a computer stays constant from one moment to the next.

    get_true_rnd, however, exploits a fundamentally chaotic property of your
    motherboard, which is otherwise digital and perfect.  Your system timer
    runs at 1.19318 MHz.  Your CPU might run at 8, 25, 33, 90, 100 etc. MHz.
    None of these numbers is an integer multiple of 1.19318.  This implies
    that, because of the way quartz oscillators are set-up electronically, your
    timer and your CPU are marching to different drum beats.  Even though these
    square wave oscillators are out of phase with one another, this routine
    would always produce the same number at the same time of day -- except for
    the fact that all physical systems involve error.  Therefore, the exact
    temperature dependence of each of these oscillators comes into question,
    and their pulses add up in a different way each time this routine is run --
    even if you reset the timer and lock out interrupts for the duration of its
    execution!  Try it if you don't believe me.

    The LSB of the 16-bit timer count is the most important because error is
    most noticable at the finest timing resolution.  You might think, then,
    that there is no point in paying attention to the MSB, which adds little or
    no randomness to the routine.  However, I recommend keeping the second XOR
    DL,AL, as it provides a fail-safe in case, for instance, Intel chooses to
    make another CPU oscillator that is an integer multiple of 1.19318 (which
    would therefore be coupled to the timer oscillator, destroying the chaos).
    If that happened, then the MSB would at least produce different numbers at
    different times of day.  There are other reasons for using the MSB, but
    they are complex and not worth getting into here.

    By the way, the PC-XT 4.77 MHz CPU oscillator is divided by four to drive
    the timer.  Therefore, since both devices are trigged by the same pulse,
    the chaotic effects inherent in other systems will _not_ be available on
    the XT (8086 and perhaps 186).  That's why I've made this a 286-and-up
    routine.

    If you have a particularly fast CPU, you may want to boost the randomness
    input in AX, as it will take more CPU cycles for the timer/CPU chaos to
    accrue.  Experimenting with high-end machines will let you know how much
    randomness is required for your application.
  *

  push cx
  push dx

  shl ax,4
  mov cx,ax                             ;cycle 16 bits [randomness] times
  get_true_rnd_loop:
    mov al,06h                          ;latch counter 0 for word read
    cli
    out 43h,al                          ;post request to timer control port
    jmp $+2                             ;IO delay (JMP to next instruction)
    in al,40h                           ;read LSB of timer count
    xor dl,al                           ;XOR-on LSB
    jmp $+2                             ;IO delay (JMP to next instruction)
    in al,40h                           ;read MSB of timer count
    sti
    xor dl,al                           ;XOR-on MSB
    rol dx,1                            ;scroll rnd # for uniform randomness

    dec cx
  jnz short get_true_rnd_loop           ;DEC CX+JNZ is usually faster than LOOP

  mov ax,dx                             ;return ax: true random number
  pop dx
  pop cx
  ret
